package com.qmth.exam.reserve.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.qmth.boot.core.concurrent.service.ConcurrentService;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.exam.reserve.bean.Constants;
import com.qmth.exam.reserve.bean.RichTextBean;
import com.qmth.exam.reserve.bean.apply.ApplyTimePeriodResult;
import com.qmth.exam.reserve.bean.apply.ApplyVO;
import com.qmth.exam.reserve.bean.apply.TicketInfo;
import com.qmth.exam.reserve.bean.apply.TimePeriodInfo;
import com.qmth.exam.reserve.bean.applytask.CurrentApplyTaskVO;
import com.qmth.exam.reserve.bean.login.LoginUser;
import com.qmth.exam.reserve.cache.CacheConstants;
import com.qmth.exam.reserve.cache.impl.ApplyTaskCacheService;
import com.qmth.exam.reserve.dao.StudentApplyDao;
import com.qmth.exam.reserve.entity.ApplyTaskEntity;
import com.qmth.exam.reserve.entity.ExamSiteEntity;
import com.qmth.exam.reserve.entity.StudentApplyEntity;
import com.qmth.exam.reserve.entity.TimePeriodEntity;
import com.qmth.exam.reserve.enums.ApplyStatus;
import com.qmth.exam.reserve.service.*;
import com.qmth.exam.reserve.util.DateUtil;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.FastDateFormat;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.text.ParseException;
import java.util.*;
import java.util.concurrent.locks.Lock;

@Service
public class ExamReserveServiceImpl implements ExamReserveService {

    private static final Logger log = LoggerFactory.getLogger(ExamReserveServiceImpl.class);

    @Autowired
    private ApplyTaskCacheService applyTaskCacheService;

    @Autowired
    private ApplyTaskService applyTaskService;

    @Autowired
    private ExamSiteService examSiteService;

    @Autowired
    private TimePeriodService timePeriodService;

    @Autowired
    private StudentApplyService studentApplyService;

    @Autowired
    private ConcurrentService concurrentService;

    @Override
    @Transactional
    public void saveStudentApply(LoginUser student, Long examSiteId, Long timePeriodId) {
        if (examSiteId == null) {
            throw new StatusException("考点ID不能为空");
        }
        if (timePeriodId == null) {
            throw new StatusException("预约时段ID不能为空");
        }

        CurrentApplyTaskVO curApplyTask = applyTaskCacheService.currentApplyTask(student.getOrgId());
        if (curApplyTask == null) {
            log.warn("预约失败，尚未开启预约任务！stuApplyTaskId:{}", student.getApplyTaskId());
            throw new StatusException("尚未开启预约任务");
        }

        if (!curApplyTask.getTaskId().equals(student.getApplyTaskId())) {
            log.warn("预约失败，当前预约任务与学生的不匹配！curApplyTaskId:{} stuApplyTaskId:{}", curApplyTask.getTaskId(), student.getApplyTaskId());
            throw new StatusException("当前预约任务与学生的不匹配");
        }

        TimePeriodEntity timePeriod = timePeriodService.getById(timePeriodId);
        if (timePeriod == null || !curApplyTask.getTaskId().equals(timePeriod.getApplyTaskId())) {
            log.warn("预约失败，当前预约时段信息有误！timePeriodId:{}", timePeriodId);
            throw new StatusException("当前预约时段信息有误");
        }

        // 考前N天，禁止考生自主预约
        Date allowDate = DateUtil.changeDateAndTimeEnd(new Date(), curApplyTask.getAllowApplyDays());
        Date curDate = new Date(timePeriod.getStartTime());
        if (curDate.before(allowDate)) {
            // 跳过过期时段，“当前时段开始时间”在“允许预约时间”之前，则禁止预约
            log.warn("预约失败，当前预约时段已禁止预约！allowApplyDays:{} timePeriodStart:{}",
                    curApplyTask.getAllowApplyDays(), timePeriod.getStartTime());
            throw new StatusException("当前预约时段已禁止预约");
        }

        // 考生预约操作锁
        String studentApplyLockKey = String.format(CacheConstants.LOCK_STUDENT_APPLY, student.getId());
        Lock studentApplyLock = concurrentService.getLock(studentApplyLockKey);
        try {
            if (!studentApplyLock.tryLock()) {
                log.warn("获取锁失败，不允许同一个考生同时操作预约！lockKey:{}", studentApplyLockKey);
                throw new StatusException(Constants.SYSTEM_BUSY);
            }

            if (concurrentService.isLocked(CacheConstants.LOCK_AUTO_APPLY)) {
                // 系统自动预约“任务执行期间”不允许预约
                log.warn("系统自动预约中，不允许考生操作预约！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
                throw new StatusException(Constants.SYSTEM_BUSY);
            }

            // 考生剩余预约次数 = 允许预约次数 - 已完成预约次数
            int studentApplyNumber = applyTaskCacheService.getStudentApplyNumber(student.getId());
            int studentApplyFinishCount = this.getStudentApplyFinishCount(student.getId());
            int unApplyNumber = studentApplyNumber - studentApplyFinishCount;
            if (unApplyNumber < 1) {
                String msg = "当前学生无剩余可约时段次数，已完成预约" + studentApplyNumber + "次";
                log.warn(msg);
                throw new StatusException(msg);
            }

            // 当前预约时段剩余可约数量
            int availableCount = this.getApplyAvailableCount(examSiteId, timePeriod.getId());
            if (availableCount < 1) {
                log.warn("预约失败，当前预约时段已约满！examSiteId:{} timePeriodId:{}", examSiteId, timePeriod.getId());
                throw new StatusException("当前预约时段已约满");
            }

            // 某考点某时段的“已预约数量” 累加1（优先执行抢占1个数量）
            applyTaskCacheService.increaseApplyFinishCount(examSiteId, timePeriodId);

            LambdaQueryWrapper<StudentApplyEntity> wrapper = new LambdaQueryWrapper<>();
            wrapper.eq(StudentApplyEntity::getStudentId, student.getId());
            wrapper.eq(StudentApplyEntity::getExamSiteId, examSiteId);
            wrapper.eq(StudentApplyEntity::getTimePeriodId, timePeriodId);
            StudentApplyEntity existEntity = studentApplyService.getOne(wrapper);
            if (existEntity != null) {
                if (!existEntity.getCancel()) {
                    // 某考点某时段的“已预约数量” 累减1（释放1个数量）
                    applyTaskCacheService.decreaseApplyFinishCount(examSiteId, timePeriodId);

                    log.warn("预约失败，当前时段已预约，请勿重复预约！applyId:{}", existEntity.getId());
                    throw new StatusException("当前时段已预约，请勿重复预约");
                }

                // 存在“已取消”预约记录，则恢复预约
                this.updateStudentApplyForCancel(existEntity.getId(), false, student.getId());
            } else {
                // 不存在则新增预约记录
                StudentApplyEntity entity = new StudentApplyEntity();
                entity.setStudentId(student.getId());
                entity.setExamSiteId(examSiteId);
                entity.setTimePeriodId(timePeriodId);
                entity.setOperateId(student.getId());
                entity.setCancel(false);
                studentApplyService.save(entity);
                log.warn("新增考生预约记录！applyId:{} examSiteId:{} timePeriodId:{}", entity.getId(), examSiteId, timePeriodId);
            }
        } finally {
            studentApplyLock.unlock();
        }
    }

    @Override
    @Transactional
    public void cancelStudentApply(LoginUser student, Long applyId) {
        if (applyId == null) {
            throw new StatusException("预约记录ID不能为空");
        }

        CurrentApplyTaskVO curApplyTask = applyTaskCacheService.currentApplyTask(student.getOrgId());
        if (curApplyTask == null) {
            log.warn("取消预约失败，尚未开启预约任务！stuApplyTaskId:{} applyId:{}", student.getApplyTaskId(), applyId);
            return;
        }
        if (!curApplyTask.getTaskId().equals(student.getApplyTaskId())) {
            log.warn("取消预约失败，当前预约任务与学生的不匹配！curApplyTaskId:{} stuApplyTaskId:{}", curApplyTask.getTaskId(), student.getApplyTaskId());
            return;
        }

        StudentApplyEntity studentApply = studentApplyService.getById(applyId);
        if (studentApply == null) {
            log.warn("取消预约失败，预约记录不存在！applyId:{}", applyId);
            return;
        }
        if (!studentApply.getStudentId().equals(student.getId())) {
            log.warn("取消预约失败，预约记录与当前学生的不匹配！applyId:{} {}!={}", applyId, studentApply.getStudentId(), student.getId());
            return;
        }

        if (studentApply.getCancel()) {
            log.warn("当前时段已取消预约，忽略重复取消！applyId:{}", applyId);
            return;
        }

        TimePeriodEntity timePeriod = timePeriodService.getById(studentApply.getTimePeriodId());
        if (timePeriod == null) {
            log.warn("取消预约失败，预约时段不存在！applyId:{} timePeriodId:{}", applyId, studentApply.getTimePeriodId());
            return;
        }

        // 考前N天，禁止考生自主取消预约
        Date allowDate = DateUtil.changeDateAndTimeEnd(new Date(), curApplyTask.getAllowApplyCancelDays());
        Date curDate = new Date(timePeriod.getStartTime());
        if (curDate.before(allowDate)) {
            // “当前时段开始时间”在“允许取消时间”之前，禁止取消预约
            log.warn("取消预约失败，当前预约记录已禁止取消预约！allowApplyCancelDays:{} timePeriodStart:{}",
                    curApplyTask.getAllowApplyCancelDays(), timePeriod.getStartTime());
            throw new StatusException("当前预约记录已禁止取消预约");
        }

        // 考生预约操作锁
        String studentApplyLockKey = String.format(CacheConstants.LOCK_STUDENT_APPLY, student.getId());
        Lock studentApplyLock = concurrentService.getLock(studentApplyLockKey);
        try {
            if (!studentApplyLock.tryLock()) {
                log.warn("获取锁失败，不允许同一个考生同时操作预约！lockKey:{}", studentApplyLockKey);
                throw new StatusException(Constants.SYSTEM_BUSY);
            }

            if (concurrentService.isLocked(CacheConstants.LOCK_AUTO_APPLY)) {
                // 系统自动预约“任务执行期间”不允许取消预约
                log.warn("系统自动预约中，不允许考生操作预约！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
                throw new StatusException(Constants.SYSTEM_BUSY);
            }

            this.updateStudentApplyForCancel(applyId, true, student.getId());

            // 某考点某时段的“已预约数量” 累减1
            applyTaskCacheService.decreaseApplyFinishCount(studentApply.getExamSiteId(), studentApply.getTimePeriodId());

            // todo 取消预约事件
        } finally {
            studentApplyLock.unlock();
        }
    }

    private void updateStudentApplyForCancel(Long applyId, boolean cancel, Long operateId) {
        LambdaUpdateWrapper<StudentApplyEntity> updateWrapper = new LambdaUpdateWrapper<>();
        updateWrapper.set(StudentApplyEntity::getCancel, cancel);
        updateWrapper.set(StudentApplyEntity::getOperateId, operateId);
        updateWrapper.set(StudentApplyEntity::getUpdateTime, System.currentTimeMillis());
        updateWrapper.eq(StudentApplyEntity::getId, applyId);
        studentApplyService.update(updateWrapper);
        log.warn("更新考生预约记录！applyId:{} cancel:{}", applyId, cancel);
    }

    private int getStudentApplyFinishCount(Long studentId) {
        LambdaQueryWrapper<StudentApplyEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(StudentApplyEntity::getStudentId, studentId);
        wrapper.eq(StudentApplyEntity::getCancel, Boolean.FALSE);
        return studentApplyService.count(wrapper);
    }

    private int getApplyAvailableCount(Long examSiteId, Long timePeriodId) {
        // 当前预约时段剩余可约数量
        int totalCount = applyTaskCacheService.getApplyTotalCount(examSiteId);
        int finishCount = applyTaskCacheService.getApplyFinishCount(examSiteId, timePeriodId);
        int availableCount = totalCount - finishCount;
        return Math.max(availableCount, 0);
    }

    @Override
    public List<ApplyVO> getStudentApplyList(LoginUser student, Boolean cancel) {
        CurrentApplyTaskVO curApplyTask = applyTaskCacheService.currentApplyTask(student.getOrgId());
        if (curApplyTask == null) {
            return new ArrayList<>();
        }

        StudentApplyDao baseMapper = (StudentApplyDao) studentApplyService.getBaseMapper();
        List<ApplyVO> list = baseMapper.getStudentApplyList(student.getId(), cancel);
        if (CollectionUtils.isEmpty(list)) {
            return new ArrayList<>();
        }

        // 考前N天，禁止考生自主取消预约
        Date allowDate = DateUtil.changeDateAndTimeEnd(new Date(), curApplyTask.getAllowApplyCancelDays());

        for (ApplyVO vo : list) {
            vo.setShowTicket(false);
            vo.setAllowCancel(false);

            if (StringUtils.isNotEmpty(vo.getTicketNumber())) {
                // 准考证号已生成，则可查看
                vo.setShowTicket(true);
            }

            Date curDate = new Date(vo.getTimePeriodStart());
            if (curDate.after(allowDate)) {
                // “当前时段开始时间”在“允许取消时间”之后，可以取消预约
                vo.setAllowCancel(true);
            }
        }

        return list;
    }

    @Override
    public List<ApplyVO> getStudentApplyListForCurrent(LoginUser student) {
        // 默认过滤掉“已取消预约”的记录
        List<ApplyVO> list = this.getStudentApplyList(student, false);

        Date now = new Date();
        List<ApplyVO> newList = new ArrayList<>();
        for (ApplyVO vo : list) {
            Date curDate = new Date(vo.getTimePeriodEnd());
            if (curDate.before(now)) {
                // “当前时段结束时间”在“当前时间”之前，则过滤掉
                continue;
            }
            newList.add(vo);
        }

        return newList;
    }

    @Override
    public TicketInfo getApplyTicket(Long studentId, Long applyId) {
        StudentApplyDao baseMapper = (StudentApplyDao) studentApplyService.getBaseMapper();
        TicketInfo info = baseMapper.getStudentApplyTicket(applyId);
        if (info == null || info.getCancel()) {
            throw new StatusException("准考证信息不存在");
        }
        if (!info.getStudentId().equals(studentId)) {
            throw new StatusException("准考证信息不匹配");
        }
        if (StringUtils.isEmpty(info.getTicketNumber())) {
            throw new StatusException("准考证信息尚未生成");
        }
        return info;
    }

    @Override
    public List<String> getApplyDateList(LoginUser student) {
        // 日期格式：yyyyMMdd
        List<String> dateList = new ArrayList<>();

        CurrentApplyTaskVO curApplyTask = applyTaskCacheService.currentApplyTask(student.getOrgId());
        if (curApplyTask == null) {
            return dateList;
        }

        LambdaQueryWrapper<TimePeriodEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.select(TimePeriodEntity::getStartTime);// 只查询startTime等字段
        wrapper.eq(TimePeriodEntity::getApplyTaskId, curApplyTask.getTaskId());
        List<TimePeriodEntity> timePeriods = timePeriodService.list(wrapper);
        if (CollectionUtils.isEmpty(timePeriods)) {
            return dateList;
        }

        // 考前N天，禁止考生自主预约
        Date allowDate = DateUtil.changeDateAndTimeEnd(new Date(), curApplyTask.getAllowApplyDays());
        Set<String> dates = new HashSet<>();
        FastDateFormat fdf = FastDateFormat.getInstance("yyyyMMdd");
        for (TimePeriodEntity timePeriod : timePeriods) {
            Date curDate = new Date(timePeriod.getStartTime());
            if (curDate.before(allowDate)) {
                // 跳过过期时段，“当前时段开始时间”在“允许预约时间”之前，则禁止预约
                continue;
            }

            dates.add(fdf.format(curDate));
        }

        dateList.addAll(dates);
        Collections.sort(dateList);
        return dateList;
    }

    @Override
    public ApplyTimePeriodResult getApplyTimePeriodList(LoginUser student, Long examSiteId, String date) {
        if (examSiteId == null) {
            throw new StatusException("考点ID不能为空");
        }
        if (StringUtils.isEmpty(date)) {
            throw new StatusException("预约日期不能为空");
        }

        Date startTime, endTime;
        try {
            FastDateFormat fdf = FastDateFormat.getInstance("yyyyMMddHHmmss");
            startTime = fdf.parse(date + "000000");
            endTime = fdf.parse(date + "235959");
        } catch (ParseException e) {
            log.error(e.getMessage(), e);
            throw new StatusException("预约日期值的格式不正确");
        }

        CurrentApplyTaskVO curApplyTask = applyTaskCacheService.currentApplyTask(student.getOrgId());
        if (curApplyTask == null) {
            throw new StatusException("尚未开启预约任务");
        }

        LambdaQueryWrapper<TimePeriodEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.select(TimePeriodEntity::getId, TimePeriodEntity::getStartTime, TimePeriodEntity::getEndTime);
        wrapper.eq(TimePeriodEntity::getApplyTaskId, curApplyTask.getTaskId());
        wrapper.between(TimePeriodEntity::getStartTime, startTime.getTime(), endTime.getTime());
        List<TimePeriodEntity> timePeriods = timePeriodService.list(wrapper);

        List<TimePeriodInfo> periodList = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(timePeriods)) {
            for (TimePeriodEntity timePeriod : timePeriods) {
                TimePeriodInfo info = new TimePeriodInfo();
                info.setTimePeriodId(timePeriod.getId());
                info.setTimePeriodStart(timePeriod.getStartTime());
                info.setTimePeriodEnd(timePeriod.getEndTime());

                int availableCount = this.getApplyAvailableCount(examSiteId, timePeriod.getId());
                info.setAvailableCount(availableCount);
                if (availableCount < 1) {
                    info.setStatus(ApplyStatus.FULL);
                } else {
                    info.setStatus(ApplyStatus.AVAILABLE);
                }

                // todo 已约

                periodList.add(info);
            }
        }

        // 考生剩余预约次数 = 允许预约次数 - 已完成预约次数
        int studentApplyNumber = applyTaskCacheService.getStudentApplyNumber(student.getId());
        int studentApplyFinishCount = this.getStudentApplyFinishCount(student.getId());
        int unApplyNumber = studentApplyNumber - studentApplyFinishCount;

        ApplyTimePeriodResult result = new ApplyTimePeriodResult();
        result.setUnApplyNumber(Math.max(unApplyNumber, 0));
        result.setPeriodList(periodList);

        return result;
    }

    @Override
    public RichTextBean getExamNotice(Long applyTaskId) {
        if (applyTaskId == null) {
            throw new StatusException("预约任务ID不能为空");
        }

        ApplyTaskEntity applyTask = applyTaskService.getById(applyTaskId);
        if (applyTask == null) {
            throw new StatusException("考试须知信息不存在");
        }

        RichTextBean bean = new RichTextBean();
        bean.setContent(applyTask.getNotice());
        return bean;
    }

    @Override
    public RichTextBean getExamSiteGuide(Long examSiteId) {
        if (examSiteId == null) {
            throw new StatusException("考点ID不能为空");
        }

        ExamSiteEntity examSite = examSiteService.getById(examSiteId);
        if (examSite == null) {
            throw new StatusException("考场引导信息不存在");
        }

        RichTextBean bean = new RichTextBean();
        bean.setContent(examSite.getGuide());
        return bean;
    }

}
